/************************************************************************
 *
 *  File: wavconverter.c
 *  Description: Convert Wav to NIOS-II Assembly format
 *
 *  Original Version
 *  Date: March 4th, 2007
 *  Author: Johnny Ho
 *
 *  Updated:
 *  March 18, 2007, Guy Lemieux: refactored code, more robust
 *  March 5, 2008, Guy Lemieux: dump halfword data only (no labels)
 *
 *  WAVE file description here:
 *  http://ccrma.stanford.edu/courses/422/projects/WaveFormat/
 *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#define MAX_LENGTH 1024

#define RETURN_EXTRA_DATA 	14
#define RETURN_NOT_ENOUGH_DATA 	13
#define RETURN_NOT_DATASIZE 12
#define RETURN_NOT_DATA 	11
#define RETURN_NOT_ALIGNED 	10
#define RETURN_NOT_STEREO 	9
#define RETURN_NOT_176400 	8
#define RETURN_NOT_44100 	7
#define RETURN_NOT_PCMUNC 	6
#define RETURN_NOT_PCM 	5
#define RETURN_NOT_FMT 	4
#define RETURN_NOT_WAVE	3
#define RETURN_NOT_RIFF	2
#define RETURN_ERROR	1
#define RETURN_OK		0
#define RETURN_DONE		-1


FILE *fp_rd;
FILE *fp_wr;

int file_bytes_remaining = 0 ;


int get_word( FILE *fp )
{
	int byte0, byte1, byte2, byte3;
	byte0 = fgetc(fp);
	byte1 = fgetc(fp);
	byte2 = fgetc(fp);
	byte3 = fgetc(fp);
	return (byte3<<24) | (byte2<<16) | (byte1<<8) | byte0;
}

int get_halfword( FILE *fp )
{
	int byte0, byte1;
	byte0 = fgetc(fp);
	byte1 = fgetc(fp);
	return (byte1<<8) | byte0;
}

int rev_word( unsigned int word )
{
	int byte0, byte1, byte2, byte3;
	byte0 = (word >> 24) & 0xff;
	byte1 = (word >> 16) & 0xff;
	byte2 = (word >>  8) & 0xff;
	byte3 = (word      ) & 0xff;
	return (byte3<<24) | (byte2<<16) | (byte1<<8) | byte0;
}

void print_error_message( int error, int error_value )
{
	switch ( error ) {
	case RETURN_ERROR:
		fprintf(stderr, "Unknown error.\n" );
		break;
	case RETURN_NOT_RIFF:
		fprintf(stderr, "Did not find string \"RIFF\" in file, found %4x.\n", error_value );
		break;
	case RETURN_NOT_WAVE:
		fprintf(stderr, "Did not find string \"WAVE\" in file, found %4x.\n", error_value );
		break;
	case RETURN_NOT_FMT:
		fprintf(stderr, "Did not find string \"fmt \" in file, found %4x.\n", error_value );
		break;
	case RETURN_NOT_PCM:
		fprintf(stderr, "Did not find value 1 for proper PCM format, found %d.\n", error_value );
		break;
	case RETURN_NOT_PCMUNC:
		fprintf(stderr, "Did not find value 1 for proper uncompressed PCM format, found %d.\n",
			error_value );
		break;
	case RETURN_NOT_44100:
		fprintf(stderr, "Did not find correct sample rate of 44100kHz, found %d.\n", error_value );
		break;
	case RETURN_NOT_176400:
		fprintf(stderr, "Did not find correct byte rate of 176400 bytes/s, found %d.\n", error_value );
		break;
	case RETURN_NOT_STEREO:
		fprintf(stderr, "Did not find 2 channels, found %d.\n", error_value );
		break;
	case RETURN_NOT_ALIGNED:
		fprintf(stderr, "Did not find word-aligned data value 4, found %d.\n", error_value );
		break;
	case RETURN_NOT_DATA:
		fprintf(stderr, "Did not find string \"DATA\" in file, found %4x.\n", error_value );
		break;
	case RETURN_NOT_DATASIZE:
		fprintf(stderr, "Inconsistency in internal file data sizes.  Found %d expected %d bytes.\n",
			error_value, file_bytes_remaining );
		break;
	case RETURN_NOT_ENOUGH_DATA:
		fprintf(stderr, "Reached end of file, but expected %d more bytes.\n", error_value );
		break;
	case RETURN_EXTRA_DATA:
		fprintf(stderr, "Reached end of sound, but more %d bytes still left in file.\n", error_value );
		break;
	case RETURN_OK:
	case RETURN_DONE:
		break;
	default:
		fprintf(stderr, "Default case, unknown RETURN_VALUE of %d.\n", error );
		break;
	}

}


int get_format_section( int *pbytes_remaining, int *perror_value )
{
	int RETURN_VALUE = RETURN_OK;
	int word;

	/* AudioFormat */
	/* Next 2 bytes should be value 1 if WAV is in proper uncompressed PCM format */
	word = get_halfword(fp_rd);
	if( word != 1 ) {
		RETURN_VALUE = RETURN_NOT_PCMUNC;
		*perror_value = word;
		goto CLEANUP;
	}
	*pbytes_remaining -= 2;

	/* NumChannels */
	/* Next 2 bytes should be value 2 if WAV is in stereo */
	word = get_halfword(fp_rd);
	if( word != 2 ) {
		RETURN_VALUE = RETURN_NOT_STEREO;
		*perror_value = word;
		goto CLEANUP;
	}
	*pbytes_remaining -= 2;

	/* SampleRate */
	/* Next 4 bytes should be value 44100 if WAV is sampled at proper rate */
	word = get_word(fp_rd);
	if( word != 44100 ) {
		RETURN_VALUE = RETURN_NOT_44100;
		*perror_value = word;
		goto CLEANUP;
	}
	*pbytes_remaining -= 4;

	/* ByteRate */
	/* Next 4 bytes should be value 44100*2*2=176400 for proper byte rate */
	word = get_word(fp_rd);
	if( word != 176400 ) {
		RETURN_VALUE = RETURN_NOT_176400;
		*perror_value = word;
		goto CLEANUP;
	}
	*pbytes_remaining -= 4;

	/* BlockAlign */
	/* Next 2 bytes should be value 2*2=4 for proper byte alignment */
	word = get_halfword(fp_rd);
	if( word != 4 ) {
		RETURN_VALUE = RETURN_NOT_ALIGNED;
		*perror_value = word;
		goto CLEANUP;
	}
	*pbytes_remaining -= 2;

	/* BitsPerSample */
	/* Next 2 bytes should be value 16 for proper sample size */
	word = get_halfword(fp_rd);
	if( word != 16 ) {
		RETURN_VALUE = RETURN_NOT_ALIGNED;
		*perror_value = word;
		goto CLEANUP;
	}
	*pbytes_remaining -= 2;

CLEANUP:
	return RETURN_VALUE;

}

int get_data_section( int *pbytes_remaining, int *perror_value )
{
	/* now, file should contain exactly bytes_remaining bytes left */

	int RETURN_VALUE = RETURN_OK;
	int byte0, byte1, byte2, byte3;

/*
	fprintf ( fp_wr, "\n.equ	WAVLEN,	%d\n", *pbytes_remaining );
	fprintf ( fp_wr,"WAV:" );
 */

	while( !feof(fp_rd) && *pbytes_remaining > 0 ) {
		/* after every pair of samples, output audio data */
		fprintf ( fp_wr, "\n.hword");
		byte0 = fgetc(fp_rd);
		byte1 = fgetc(fp_rd);
		byte2 = fgetc(fp_rd);
		byte3 = fgetc(fp_rd);
		fprintf ( fp_wr, "  0x%02x%02x", byte1, byte0 );
		fprintf ( fp_wr, ", 0x%02x%02x", byte3, byte2 );
		*pbytes_remaining -=4;
	}

	if( *pbytes_remaining < 0 ) {
		RETURN_VALUE = RETURN_NOT_ENOUGH_DATA;
		*perror_value = *pbytes_remaining;
		goto CLEANUP;
	}

	if( *pbytes_remaining > 0 ) {
		RETURN_VALUE = RETURN_EXTRA_DATA;
		*perror_value = *pbytes_remaining;
		goto CLEANUP;
	}

CLEANUP:
	return RETURN_VALUE;

}

void get_section_id_length( unsigned int *pSectionID, int *pSectionLength )
{
	/* first word is Big Endian ID */
	*pSectionID = (unsigned int)rev_word(get_word(fp_rd));

	/* second word is section length */
	*pSectionLength = get_word(fp_rd);

	fprintf( stderr, "Reading section 0x%08lx, length %d bytes.\n", *pSectionID, *pSectionLength );
}

int convert_file()
{
	unsigned int byte_count = 0;
	unsigned int word = 0;
	unsigned int sectionID = 0;
	int sectionLength = 0;

	int section_bytes_remaining = 0;

	char file_name[MAX_LENGTH];

	int RETURN_VALUE = RETURN_OK;
	int error_value = 0;

	printf(	"\nWAV file converter for EECE259.\n"
		"Please enter the file name, eg money.wav, or done to exit.\n");
	gets(file_name);

	if (	file_name == NULL 
/*		||	!feof(stdin)      */
		||	!strcmp(file_name, "done")
		|| 	!strcmp(file_name, "exit")  ) {
		RETURN_VALUE = RETURN_DONE;
		goto CLEANUP;
	}


	if ( (fp_rd = fopen(file_name, "rb")) == 0 ) {
		fprintf(stderr, "Cannot open file \"%s\" to read\n", file_name );
		fflush(stderr);
		goto CLEANUP;
	}	
	if ( (fp_wr = fopen("wavdata.s", "wb")) == 0 ) {
		fprintf(stderr, "Cannot create file wavdata.s to write\n" );
		fflush(stderr);
		goto CLEANUP;
	}

	fprintf ( stderr, "\n***************Conversion Begins***************\n");
	fprintf ( fp_wr, "/******************* WAV data *******************/\n\n");
	fprintf ( fp_wr, "/* Audio is 44.1kHz, 16b stereo *****************/\n");
	fprintf ( fp_wr, "/* Each sample consists of LEFT and RIGHT data  */\n");
	fprintf ( fp_wr, "/* The first two bytes (halfword) is LEFT data  */\n");
	fprintf ( fp_wr, "/* The next  two bytes (halfword) is RIGHT data */\n");
	fprintf ( fp_wr, "/************************************************/\n");
	fprintf ( fp_wr, "/* Notice the values do not change too quickly  */\n");
	fprintf ( fp_wr, "/* between samples. This suggests the sample    */\n");
	fprintf ( fp_wr, "/* rate is faster than the audio rate, which is */\n");
	fprintf ( fp_wr, "/* essential for good quality sound.            */\n");
	fprintf ( fp_wr, "/************************************************/\n");

	/* Parsing Input File*/

	fprintf( stderr, "Looking for section \"RIFF\".\n" );

	get_section_id_length( &sectionID, &sectionLength );

	/* First sectionId should be letters "RIFF" */
	if( sectionID != 0x52494646 ) {
		RETURN_VALUE = RETURN_NOT_RIFF;
		error_value = word;
		goto CLEANUP;
	}
	file_bytes_remaining = sectionLength;

	fprintf( stderr, "Found section \"RIFF\".\n" );

	/* now, file should contain exactly file_bytes_remaining bytes left */

	/* Next 4 bytes should be letters "WAVE" */
	word = rev_word(get_word(fp_rd));
	if( word != 0x57415645 ) {
		RETURN_VALUE = RETURN_NOT_WAVE;
		error_value = word;
		goto CLEANUP;
	}
	file_bytes_remaining -= 4;

	fprintf( stderr, "Found section \"WAVE\". Yes, this is a RIFF file in WAV format.\n" );

	/* File will now be composed of a number of Subchunk sections.
	 * We'll look at each sectionID, looking for "fmt " and "data".
	 * If the "fmt " section contains correct parameters,
	 * we'll translate next "data" section and quit.
	 */

	fprintf( stderr, "Looking for section \"fmt \".\n" );

	do {
		get_section_id_length( &sectionID, &sectionLength );
		file_bytes_remaining -= 8;
		section_bytes_remaining = sectionLength;

		/* Next 4 bytes should be letters "fmt " */
		if( sectionID == 0x666d7420 ) {
			RETURN_VALUE = get_format_section( &section_bytes_remaining, &error_value );
			if( RETURN_VALUE == RETURN_OK ) {
				fprintf( stderr, "Found section \"fmt \" that contains proper"
					" encoding information.\n" );
				if( section_bytes_remaining > 0 ) {
					fprintf( stderr, "Some leftover bytes in the format section;"
						" nonstandard but OK.\n" );
					while( section_bytes_remaining-- )	// should actually be 0 already
						fgetc( fp_rd );
				}
				file_bytes_remaining -= sectionLength;
				break;
			} else {
				fprintf( stderr, "Found section \"fmt \", but the data is an unknown encoding.\n" );
				print_error_message( RETURN_VALUE, error_value );
				fprintf( stderr, "Will continue looking for another section with correct encoding."
					" Skipping %d bytes.\n",  section_bytes_remaining );
				while( section_bytes_remaining-- )
					fgetc( fp_rd );
				file_bytes_remaining -= sectionLength;
			}
		} else {
			fprintf( stderr, "Found unknown section type, skipping %d bytes.\n", section_bytes_remaining );
			while( section_bytes_remaining-- > 0 )
				fgetc( fp_rd );
			file_bytes_remaining -= sectionLength;
		}

		if( file_bytes_remaining <= 0 || feof(fp_rd) )
			break;

	} while(1);

	/* Next section should be "data".
	 */

	fprintf( stderr, "Looking for section \"data\".\n" );

	do {
		get_section_id_length( &sectionID, &sectionLength );
		file_bytes_remaining -= 8;
		section_bytes_remaining = sectionLength;

		if( sectionID == 0x64617461 ) {
			fprintf( stderr, "Found section \"data\". Extracting %d kB music data.\n",
				section_bytes_remaining/1024 );
			RETURN_VALUE = get_data_section( &section_bytes_remaining, &error_value );
			file_bytes_remaining -= sectionLength;
			if( sectionLength > 1024*1024 ) {
				fprintf( stderr, "WARNING: this music data is over 1024 kB in size.\n"
					"Remember, the DE board has only 8192 kB total.\n", sectionLength/1024 );
			}
			break;
		} else {
			fprintf( stderr, "Found unknown section type, skipping. Looking for next data section.\n" );
			while( section_bytes_remaining-- )
				fgetc( fp_rd );
			file_bytes_remaining -= sectionLength;
		}

		if( file_bytes_remaining <= 0 || feof(fp_rd) )
			break;

	} while(1);


CLEANUP:
	if( RETURN_VALUE != RETURN_DONE ) {
		fprintf ( fp_wr, "\n\n");
		printf ("\n***************Conversion done***************\n");
		printf ("\n*************** see wavdata.s ***************\n");
		fclose(fp_wr);
		fclose(fp_rd);

		print_error_message(RETURN_VALUE, error_value);
	}

	return RETURN_VALUE;
}


int main (int argc, char *argv )
{
	int rc;
	do {
		rc = convert_file();
	} while( rc != RETURN_DONE );
}